"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ValidationService = void 0;
const web3_1 = __importDefault(require("web3"));
const core_1 = require("@ethereum-sourcify/core");
const adm_zip_1 = __importDefault(require("adm-zip"));
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
/**
 * Regular expression matching metadata nested within another json.
 */
const NESTED_METADATA_REGEX = /"{\\"compiler\\":{\\"version\\".*?},\\"version\\":1}"/;
const CONTENT_VARIATORS = [
    (content) => content,
    (content) => content.replace(/\r?\n/g, "\r\n"),
    (content) => content.replace(/\r\n/g, "\n")
];
const ENDING_VARIATORS = [
    (content) => content,
    (content) => content.trimEnd(),
    (content) => content.trimEnd() + "\n",
    (content) => content.trimEnd() + "\r\n",
    (content) => content + "\n",
    (content) => content + "\r\n"
];
class ValidationService {
    /**
     * @param logger a custom logger that logs all errors; undefined or no logger provided turns the logging off
     */
    constructor(logger) {
        this.logger = logger;
    }
    checkPaths(paths, ignoring) {
        const files = [];
        paths.forEach(path => {
            if (fs_1.default.existsSync(path)) {
                this.traversePathRecursively(path, filePath => {
                    const fullPath = path_1.default.resolve(filePath);
                    const file = { buffer: fs_1.default.readFileSync(filePath), path: fullPath };
                    files.push(file);
                });
            }
            else if (ignoring) {
                ignoring.push(path);
            }
        });
        return this.checkFiles(files);
    }
    checkFiles(files, unused) {
        const inputFiles = this.findInputFiles(files);
        const parsedFiles = inputFiles.map(pathBuffer => ({ content: pathBuffer.buffer.toString(), path: pathBuffer.path }));
        const { metadataFiles, sourceFiles } = this.splitFiles(parsedFiles);
        const checkedContracts = [];
        const errorMsgMaterial = [];
        const byHash = this.storeByHash(sourceFiles);
        const usedFiles = [];
        metadataFiles.forEach(metadata => {
            const { foundSources, missingSources, invalidSources, metadata2provided } = this.rearrangeSources(metadata, byHash);
            const currentUsedFiles = Object.values(metadata2provided);
            usedFiles.push(...currentUsedFiles);
            const checkedContract = new core_1.CheckedContract(metadata, foundSources, missingSources, invalidSources);
            checkedContracts.push(checkedContract);
            if (!core_1.CheckedContract.isValid(checkedContract)) {
                errorMsgMaterial.push(checkedContract.getInfo());
            }
        });
        if (errorMsgMaterial.length) {
            const msg = errorMsgMaterial.join("\n");
            if (this.logger)
                this.logger.error(msg);
        }
        if (unused) {
            this.extractUnused(sourceFiles, usedFiles, unused);
        }
        return checkedContracts;
    }
    /**
     * Traverses the given files, unzipping any zip archive.
     *
     * @param files the array containing the files to be checked
     * @returns an array containing the provided files, with any zips being unzipped and returned
     */
    findInputFiles(files) {
        const inputFiles = [];
        for (const file of files) {
            if (this.isZip(file.buffer)) {
                this.unzip(file, files);
            }
            else {
                inputFiles.push(file);
            }
        }
        return inputFiles;
    }
    /**
     * Checks whether the provided file is in fact zipped.
     * @param file the buffered file to be checked
     * @returns true if the file is zipped; false otherwise
     */
    isZip(file) {
        try {
            new adm_zip_1.default(file);
            return true;
        }
        catch (err) {
            undefined;
        }
        return false;
    }
    /**
     * Unzips the provided file buffer to the provided array.
     *
     * @param zippedFile the buffer containin the zipped file to be unpacked
     * @param files the array to be filled with the content of the zip
     */
    unzip(zippedFile, files) {
        const timestamp = Date.now().toString() + "-" + Math.random().toString().slice(2);
        const tmpDir = `tmp-unzipped-${timestamp}`;
        new adm_zip_1.default(zippedFile.buffer).extractAllTo(tmpDir);
        this.traversePathRecursively(tmpDir, filePath => {
            const file = { buffer: fs_1.default.readFileSync(filePath), content: zippedFile.path };
            files.push(file);
        });
        this.traversePathRecursively(tmpDir, fs_1.default.unlinkSync, fs_1.default.rmdirSync);
    }
    /**
     * Selects metadata files from an array of files that may include sources, etc
     * @param  {string[]} files
     * @return {string[]}         metadata
     */
    splitFiles(files) {
        const metadataFiles = [];
        const sourceFiles = [];
        const malformedMetadataFiles = [];
        for (const file of files) {
            let metadata = this.extractMetadataFromString(file.content);
            if (!metadata) {
                const matchRes = file.content.match(NESTED_METADATA_REGEX);
                if (matchRes) {
                    metadata = this.extractMetadataFromString(matchRes[0]);
                }
            }
            if (metadata) {
                try {
                    this.assertObjectSize(metadata.settings.compilationTarget, 1);
                    metadataFiles.push(metadata);
                }
                catch (err) {
                    malformedMetadataFiles.push(file.path);
                }
            }
            else {
                sourceFiles.push(file);
            }
        }
        let msg = "";
        if (malformedMetadataFiles.length) {
            const responsibleFiles = malformedMetadataFiles.every(Boolean) ?
                malformedMetadataFiles.join(", ") : `${malformedMetadataFiles.length} metadata files`;
            msg = `Malformed settings.compilationTarget in: ${responsibleFiles}`;
        }
        else if (!metadataFiles.length) {
            msg = "Metadata file not found. Did you include \"metadata.json\"?";
        }
        if (msg) {
            if (this.logger)
                this.logger.error(msg);
            throw new Error(msg);
        }
        return { metadataFiles, sourceFiles };
    }
    /**
     * Validates metadata content keccak hashes for all files and
     * returns mapping of file contents by file name
     * @param  {any}       metadata
     * @param  {Map<string, any>}  byHash    Map from keccak to source
     * @return foundSources, missingSources, invalidSources
     */
    rearrangeSources(metadata, byHash) {
        const foundSources = {};
        const missingSources = {};
        const invalidSources = {};
        const metadata2provided = {}; // maps fileName as in metadata to the fileName of the provided file
        for (const fileName in metadata.sources) {
            const sourceInfo = metadata.sources[fileName];
            let file = { content: undefined };
            file.content = sourceInfo.content;
            const hash = sourceInfo.keccak256;
            if (file.content) {
                if (web3_1.default.utils.keccak256(file.content) != hash) {
                    const msg = "The calculated and the provided hash values don't match.";
                    invalidSources[fileName] = msg;
                    continue;
                }
            }
            else {
                const pathContent = byHash.get(hash);
                if (pathContent) {
                    file = pathContent;
                    metadata2provided[fileName] = pathContent.path;
                } // else: no file has the hash that was searched for
            }
            if (file && file.content) {
                foundSources[fileName] = file;
            }
            else {
                missingSources[fileName] = { keccak256: hash, urls: sourceInfo.urls };
            }
        }
        return { foundSources, missingSources, invalidSources, metadata2provided };
    }
    /**
     * Generates a map of files indexed by the keccak hash of their content.
     *
     * @param  {string[]}  files Array containing sources.
     * @returns Map object that maps hash to PathContent.
     */
    storeByHash(files) {
        const byHash = new Map();
        for (const pathContent of files) {
            for (const variation of this.generateVariations(pathContent)) {
                const calculatedHash = web3_1.default.utils.keccak256(variation.content);
                byHash.set(calculatedHash, variation);
            }
        }
        return byHash;
    }
    generateVariations(pathContent) {
        const variations = [];
        const original = pathContent.content;
        for (const contentVariator of CONTENT_VARIATORS) {
            const variatedContent = contentVariator(original);
            for (const endingVariator of ENDING_VARIATORS) {
                const variation = endingVariator(variatedContent);
                variations.push(variation);
            }
        }
        return variations.map(content => {
            return { content, path: pathContent.path };
        });
    }
    extractUnused(inputFiles, usedFiles, unused) {
        const usedFilesSet = new Set(usedFiles);
        const tmpUnused = inputFiles.map(pc => pc.path).filter(file => !usedFilesSet.has(file));
        unused.push(...tmpUnused);
    }
    extractMetadataFromString(file) {
        try {
            let obj = JSON.parse(file);
            if (this.isMetadata(obj)) {
                return obj;
            }
            // if the input string originates from a file where it was double encoded (e.g. truffle)
            obj = JSON.parse(obj);
            if (this.isMetadata(obj)) {
                return obj;
            }
        }
        catch (err) {
            undefined;
        }
        return null;
    }
    /**
     * A method that checks if the provided object was generated as a metadata file of a Solidity contract.
     * Current implementation is rather simplistic and may require further engineering.
     *
     * @param metadata the JSON to be checked
     * @returns true if the provided object is a Solidity metadata file; false otherwise
     */
    isMetadata(obj) {
        return (obj.language === "Solidity") &&
            !!obj.compiler;
    }
    /**
     * Applies the provided worker function to the provided path recursively.
     *
     * @param path the path to be traversed
     * @param worker the function to be applied on each file that is not a directory
     * @param afterDir the function to be applied on the directory after traversing its children
     */
    traversePathRecursively(path, worker, afterDirectory) {
        if (!fs_1.default.existsSync(path)) {
            const msg = `Encountered a nonexistent path: ${path}`;
            if (this.logger) {
                this.logger.error(msg);
            }
            throw new Error(msg);
        }
        const fileStat = fs_1.default.lstatSync(path);
        if (fileStat.isFile()) {
            worker(path);
        }
        else if (fileStat.isDirectory()) {
            fs_1.default.readdirSync(path).forEach(nestedName => {
                const nestedPath = path_1.default.join(path, nestedName);
                this.traversePathRecursively(nestedPath, worker, afterDirectory);
            });
            if (afterDirectory) {
                afterDirectory(path);
            }
        }
    }
    /**
     * Asserts that the number of keys of the provided object is expectedSize.
     * If not, logs an appropriate message (if log function provided) and throws an Error.
     * @param object the object to check
     * @param expectedSize the size that the object should have
     */
    assertObjectSize(object, expectedSize) {
        let err = "";
        if (!object) {
            err = `Cannot assert for ${object}.`;
        }
        else {
            const objectSize = Object.keys(object).length;
            if (objectSize !== expectedSize) {
                err = `Error in size assertion! Actual size: ${objectSize}. Expected size: ${expectedSize}.`;
            }
        }
        if (err) {
            if (this.logger) {
                this.logger.error({ loc: "[VALIDATION:SIZE_ASSERTION]" }, err);
            }
            throw new Error(err);
        }
    }
}
exports.ValidationService = ValidationService;
//# sourceMappingURL=ValidationService.js.map